from datetime import datetime
from pathlib import Path
import os
import re

import pandas as pd
from tqdm import tqdm


def get_logs_df(path: Path, ignore_prefix=None):
    def parse(filename: str):
        SPECIAL_JOBNAMES = ("LOCAL_SETUP", "REMOTE_SETUP", "SYNCHRONIZE")
        # e.g. o005_19880101_fc0_60_20_APP_WILDFIRES_FWI.20250521035611.err
        try:
            logname_timestamp, logkind = filename.rsplit(".", 1)
        except ValueError:
            logname_timestamp = filename
            logkind = "unknown"

        try:
            logname, timestamp = logname_timestamp.split(".", 1)
        except ValueError:
            logname = filename
            timestamp = "unknown"

        for jobname in SPECIAL_JOBNAMES:
            if jobname in logname:
                startdate = None
                ensemble = None
                chunk = 0
                split = 0
                job_jobname = [jobname]
                expid_info = logname.replace(jobname, "").strip("_")
                while "__" in expid_info:
                    expid_info = expid_info.replace("__", "_")
                if "_" in expid_info:
                    expid, info = expid_info.split("_", 1)
                else:
                    expid, info = expid_info, ""
                break
        else:
            try:
                expid, startdate, ensemble, info = logname.split("_", 3)
                chunk_split = [element for element in info.split("_") if element.isdigit()]
                job_jobname = [element for element in info.split("_") if not element.isdigit()]
        
                # Parse chunk/split
                
                match len(chunk_split):
                    case 0:
                        chunk, split = 0, 0
                    case 1:
                        chunk, split = chunk_split[0], 0
                    case 2:
                        chunk, split = chunk_split
                    case _:
                        raise RuntimeError(filename)
            except ValueError as e:
                print(logname)
                raise ValueError(filename)

        # Parse job/job name
        if len(job_jobname) == 1:
            job, jobname = job_jobname[0], None
        else:
            job, jobname = job_jobname[0], "_".join(job_jobname[1:])        

        return {
            "timestamp": timestamp,
            "logkind": logkind,
            "expid": expid,
            "startdate": startdate,
            "chunk": chunk,
            "split": split,
            "job": job,
            "jobname": jobname,
        }

    df_columns = ["timestamp", "logkind", "expid", "startdate", "chunk", "split", "job", "", "filename"]
    
    rows = []
    
    for filename in tqdm(os.listdir(path)):
        if (
            ignore_prefix is not None
            and any(filename.startswith(prefix) for prefix in ignore_prefix)
        ):
            continue
        try:
            logdct = parse(filename)
            logdct["filename"] = filename
            rows.append(logdct)
        except (IndexError, ValueError, RuntimeError) as exc:
            print(type(exc))
            print("Skipping file:", filename)
            continue
    
    # Create DataFrame once at the end
    return pd.DataFrame(rows, columns=df_columns).astype({"chunk": int, "split": int})


def get_aslogs_df(path: Path, ignore_prefix=None):
    def parse(filename: str):
        re_comments = (r"STAT_\d+", r"COMPLETED", r"TOTAL_STATS")
        # e.g. o005_19880101_fc0_39_6_APP_HYDROLAND_STAT_0
        if filename.endswith(".cmd"):
            comment = "cmd"
            filename = filename.removesuffix(".cmd")
        else:
            comment = ""
        try:
            expid, startdate, member, extra = filename.split("_", 3)
            if extra.split("_")[0].isdigit():
                chunk, extra = extra.split("_", 1)
            else:
                chunk = "0"
        except ValueError:
            raise
        parts = extra.split("_")
        if parts[0].isdigit():
            split = parts[0]
            extra = "_".join(parts[1:])
        else:
            split = -1
        for re_comment in re_comments:
            match_ = re.search(re_comment, extra)
            if match_:
                comment += extra[match_.start():match_.end()]
                job = (extra[:match_.start()] + extra[match_.end():]).strip("_")
                break
        else:
            job = extra
        return {
            "expid": expid,
            "startdate": startdate,
            "chunk": chunk,
            "split": split,
            "member": member,
            "job": job,
            "comment": comment,
        }

    df_columns = ["expid", "startdate", "member", "chunk", "split", "job", "comment", "filename"]
    
    rows = []
    
    for filename in tqdm(os.listdir(path)):
        if (
            ignore_prefix is not None
            and any(filename.startswith(prefix) for prefix in ignore_prefix)
        ):
            continue
        try:
            logdct = parse(filename)
            logdct["filename"] = filename
            rows.append(logdct)
        except (IndexError, ValueError, RuntimeError):
            print("Skipping file:", filename)
            continue
    
    # Create DataFrame once at the end
    return pd.DataFrame(rows, columns=df_columns)


def filter_logs_df(df, **kwargs):
    mask = pd.Series([True] * len(df), index=df.index)
    for key, value in kwargs.items():
        mask &= (df[key] == value)
    return df[mask].astype({"chunk": int, "split": int}).sort_values(["chunk", "split"])


def jobs_df_from_aslogs_df(df, path: Path):
    jobs = []
    df_stats = df[df["comment"] == "TOTAL_STATS"].drop(columns="comment").astype({"chunk": int, "split": int})
    for _, row in tqdm(df_stats.iterrows(), total=len(df_stats)):
        rowdct = dict(row)
        del rowdct["filename"]
        with open(path / row["filename"]) as rf:
            for i, line in enumerate(rf.read().splitlines(), start=1):
                rowdct.update()
                jobs.append({**rowdct, "num": i, **dict(zip(("submitted", "started", "finished", "status"), line.split()))})
    df_jobs = pd.DataFrame(jobs)
    for column in ("submitted", "started", "finished"):
        df_jobs[column] = df_jobs[column].apply(lambda x: datetime.strptime(x, "%Y%m%d%H%M%S"))
    return df_jobs
