"""Plotting utilities for AS analytics."""

from datetime import datetime
from pathlib import Path
from typing import Optional, Sequence

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.colors as mcolors
import pandas as pd
from tqdm import tqdm
import plotly.express as px
import plotly.graph_objects as go


def _prepare_dataframe_pivot(
    df: pd.DataFrame,
    pivot: bool | str,
    starttype: str,
) -> tuple[pd.DataFrame, list[str]]:
    """
    Prepares a DataFrame for plotting by assigning y-axis positions and optionally
    appending chunk identifiers to job names.

    Parameters:
        df (pd.DataFrame): Input DataFrame. Must contain 'job', 'chunk', 'started', 'finished'.
        pivot_chunks (bool or str): Determines how job/chunk/split breakdown is handled:
            - "split": Always
            - "chunk": Always append chunk info to job labels.
            - "on_overlap": Append chunk info only if the same job has overlapping chunk times.
            - False: Do not include chunk info, use job names only.
        starttype (str): Either "started" or "submitted".

    Returns:
        tuple:
            - pd.DataFrame: Modified DataFrame with added 'y' and possibly 'job_chunk' columns.
            - list: Sorted labels used for the y-axis.
            - str: Label for the y axis
    """
    job_chunk_labels = []
    if pivot == "on_overlap":
        pivot_jobs = {}
        for job, df_j in df.groupby("job"):
            df_j = df_j.sort_values("chunk")
            job_chunks = df_j["chunk"].unique()
            flag = False
            for c1, c2 in zip(job_chunks[:-1], job_chunks[1:]):
                df_c1 = df_j[df_j["chunk"] == c1]
                df_c2 = df_j[df_j["chunk"] == c2]
                if df_c1["finished"].max() >= df_c2[starttype].min():
                    flag = True
                    break
            pivot_jobs[job] = flag

        df["job_chunk"] = df.apply(
            lambda row: (
                f"{row['job']} (CHUNK {row['chunk']})"
                if pivot_jobs.get(row["job"], False)
                else row["job"]
            ),
            axis=1,
        )
        job_chunk_labels = sorted(df["job_chunk"].unique())
        df["y"] = df["job_chunk"].map(
            {label: i for i, label in enumerate(job_chunk_labels)}
        )
        ylabel = "JOB / CHUNK"

    elif pivot == "split":
        df["job_chunk_split"] = (
            df["job"] + " (CHUNK " + df["chunk"] + " SPLIT " + df["split"] + ")"
        )
        job_chunk_labels = sorted(df["job_chunk_split"].unique())
        df["y"] = df["job_chunk_split"].map(
            {label: i for i, label in enumerate(job_chunk_labels)}
        )
        ylabel = "JOB / CHUNK / SPLIT"

    elif pivot == "member":
        df["job_chunk_split_member"] = (
            df["job"] +  " (" + df["chunk"] + " / " + df["split"] + " / " + df["member"] + ")"
        )
        job_chunk_labels = sorted(df["job_chunk_split_member"].unique())
        df["y"] = df["job_chunk_split_member"].map(
            {label: i for i, label in enumerate(job_chunk_labels)}
        )
        ylabel = "JOB (CHUNK / SPLIT / MEMBER)"

    elif pivot == "chunk" or pivot is True:
        df["job_chunk"] = df["job"] + " (CHUNK " + df["chunk"].astype(str) + ")"
        job_chunk_labels = sorted(df["job_chunk"].unique())
        df["y"] = df["job_chunk"].map(
            {label: i for i, label in enumerate(job_chunk_labels)}
        )
        ylabel = "JOB / CHUNK"

    else:
        df["y"] = df["job"].map(
            {label: i for i, label in enumerate(sorted(df["job"].unique()))}
        )
        job_chunk_labels = sorted(df["job"].unique())
        ylabel = "JOB"

    return df, job_chunk_labels, ylabel


def plot_job_gantt_chart(
    df: pd.DataFrame,
    title: str = "Job Timeline Gantt Chart",
    color_key: str = "job",
    starttype: str = "started",
    alpha: float = 1.0,
    pivot: bool | str = "on_overlap",
    cmap: Optional[str | mcolors.Colormap] = None,
    interval: int = -1,
    add_durations: bool = False,
    edge: bool = True,
    bar_height: float = 0.4,
    height_factor: float = 0.4,
) -> plt.Figure:
    """
    Plot a Gantt chart of job execution timelines.

    Args:
        df (pd.DataFrame): DataFrame with 'job', 'started', 'finished', 'chunk'.
        title (str): Title of the chart.
        color_key(str): Column to base the color_keyon ('job', 'chunk' or 'status').
        starttype (str): Column to use for start time ('started' or 'submitted').
        alpha (float): Transparency level of the bars.
        pivot (str or bool): Whether to separate jobs by chunk. If "on_overlap",
                             separates only if chunk times overlap for a specific job.
                             If "split", separates by split also.
        cmap (str or Colormap): Matplotlib colormap name or instance.
        interval (int): X-axis tick interval in hours. -1 means automatic.
        add_durations (bool): Whether to add the durations at the end of each entry.

    Returns:
        plt.Figure: The generated figure.
    """
    df = df.copy()

    # Add duration
    df["duration"] = df["finished"] - df[starttype]

    # Normalize chunk and split labels. Ensure sorting
    df["chunk"] = (
        df["chunk"].astype(str).str.zfill(max(len(x) for x in df["chunk"].astype(str)))
    )
    df["split"] = (
        df["split"].astype(str).str.zfill(max(len(x) for x in df["split"].astype(str)))
    )

    df, job_chunk_labels, ylabel = _prepare_dataframe_pivot(df, pivot, starttype)

    # Set color mapping
    if color_key not in ("job", "chunk", "status"):
        raise ValueError(
            f"color_key must be either 'job', 'chunk', 'status'. It was '{color_key}'."
        )
    color_keys = sorted(df[color_key].unique())
    cmap = (
        plt.get_cmap(cmap) if isinstance(cmap, str) else (cmap or plt.get_cmap("tab20"))
    )
    color_map = {key: cmap(i % 20) for i, key in enumerate(color_keys)}
    df["color"] = df[color_key].map(color_map)

    # Plot
    fig, ax = plt.subplots(figsize=(12, max(4, len(job_chunk_labels) * height_factor)))
    ax.grid(which="major", axis="x", alpha=0.8)

    for _, row in tqdm(df.iterrows(), total=len(df)):
        ax.barh(
            y=row["y"],
            width=row["finished"] - row[starttype],
            left=row[starttype],
            height=bar_height,
            color=row["color"],
            edgecolor="black" if edge else None,
            linewidth=0.5,
            alpha=alpha,
            zorder=2,
        )

    # Add durations
    if add_durations:
        total_duration_by_y = df.groupby("y")["duration"].sum()
        for y_val, duration in total_duration_by_y.items():
            duration_min = int(round(duration.total_seconds(), 0)) // 60
            duration_str = f"{duration_min // 60} h {duration_min % 60} min"
            row_subset = df[df["y"] == y_val]
            bar_end = row_subset["finished"].max()
            ax.text(
                bar_end,
                y_val,
                "  " + duration_str,
                va="center",
                ha="left",
                fontsize=9,
                color="#333333",
            )

    # Y-axis
    ax.set_yticks(range(len(job_chunk_labels)))
    ax.set_yticklabels(job_chunk_labels)
    ax.set_ylabel(ylabel)
    ax.set_xlabel("Time")
    ax.set_title(title)

    # X-axis
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%b %d %H:%M"))

    if interval == -1:
        ax.xaxis.set_major_locator(mdates.AutoDateLocator())
    else:
        ax.xaxis.set_major_locator(mdates.HourLocator(interval=interval))
    fig.autofmt_xdate()

    # Legend
    handles = [
        plt.Line2D([0], [0], color=color_map[key], lw=6, label=key)
        for key in color_keys
    ]
    ax.legend(
        handles=handles,
        title=color_key.capitalize(),
        bbox_to_anchor=(1.05, 1),
        loc="upper left",
    )

    fig.tight_layout()
    return fig


def plot_job_gantt_chart_plotly(
    df: pd.DataFrame,
    title: str = "Job Timeline Gantt Chart",
    color_key: str = "job",
    starttype: str = "started",
    pivot: bool | str = "on_overlap",
    add_durations: bool = False,
    height_factor: float = 0.4,
) -> go.Figure:
    """
    Plot an interactive Gantt chart of job execution timelines using Plotly.

    Args:
        df (pd.DataFrame): DataFrame with 'job', 'started', 'finished', 'chunk'.
        title (str): Title of the chart.
        color_key (str): Column to base the color on ('job', 'chunk' or 'status').
        starttype (str): Column to use for start time ('started' or 'submitted').
        pivot (str or bool): Whether to separate jobs by chunk. If "on_overlap",
                             separates only if chunk times overlap for a specific job.
                             If "split", separates by split also.
        add_durations (bool): Whether to add the durations at the end of each entry.
        height_factor (float): Factor to determine plot height.

    Returns:
        go.Figure: The generated Plotly figure.
    """
    df = df.copy()

    # Add duration
    df["duration"] = df["finished"] - df[starttype]

    # Normalize chunk and split labels for sorting
    df["chunk"] = (
        df["chunk"].astype(str).str.zfill(max(len(x) for x in df["chunk"].astype(str)))
    )
    df["split"] = (
        df["split"].astype(str).str.zfill(max(len(x) for x in df["split"].astype(str)))
    )

    df, job_chunk_labels, _ = _prepare_dataframe_pivot(df, pivot, starttype)

    y_axis_label = "Job"
    if pivot:
        if pivot == "split":
            y_axis_label = "Job/Chunk/Split"
            df["y_label"] = df["job_chunk_split"]
        elif pivot == "member":
            y_axis_label = "Job (Chunk / Split / Member)"
            df["y_label"] = df["job_chunk_split_member"]
        else:
            y_axis_label = "Job/Chunk"
            df["y_label"] = df["job_chunk"]
    else:
        df["y_label"] = df["job"]

    fig = px.timeline(
        df,
        x_start=starttype,
        x_end="finished",
        y="y_label",
        color=color_key,
        title=title,
        hover_data=["job", "chunk", "split", "status"],
    )

    # Custom color mapping for status-based coloring
    if color_key == "status":
        # Define custom colors for different statuses
        status_colors = {
            "COMPLETED": "#1f77b4",  # Blue
            "FAILED": "#d62728",     # Red
            "RUNNING": "#2ca02c",    # Green
            "SUBMITTED": "#ff7f0e",  # Orange
        }
        
        # Update the figure with custom colors
        fig.update_traces(
            marker=dict(
                color=[status_colors.get(status, "#17a2b8") for status in df[color_key]]
            )
        )
        
        # Update legend with custom colors
        for trace in fig.data:
            if hasattr(trace, 'name') and trace.name in status_colors:
                trace.marker.color = status_colors[trace.name]

    if add_durations:
        total_durations = (
            df.groupby("y_label")["duration"].sum().dt.total_seconds() / 60
        )
        annotations = []
        for label in job_chunk_labels:
            duration_min = int(round(total_durations.get(label, 0), 0))
            duration_str = f"{duration_min // 60} h {duration_min % 60} min"
            annotations.append(
                dict(
                    xref="paper",
                    yref="y",
                    x=1.01,
                    y=label,
                    text=duration_str,
                    showarrow=False,
                    xanchor="left",
                    yanchor="middle",
                    font=dict(size=10),
                )
            )
        fig.update_layout(annotations=annotations)

    fig.update_yaxes(
        categoryorder="array",
        categoryarray=job_chunk_labels,
        title=y_axis_label,
        autorange="reversed",
        tickfont=dict(size=10),
    )  # Keep original order

    # Calculate a more reasonable height that fits the screen
    # Base height on number of jobs but cap it at a reasonable maximum
    base_height = 600
    job_height = min(len(job_chunk_labels) * 25, 800)  # Cap at 800px for many jobs
    plot_height = max(base_height, job_height)

    fig.update_layout(
        height=plot_height,
        xaxis_title="Time",
        legend_title_text=color_key.capitalize(),
        margin=dict(l=10, r=150, t=50, b=10),  # Increased right margin for legend
        font=dict(size=12),
        legend=dict(
            font=dict(size=10),
            x=1.02,  # Position legend outside the plot
            xanchor="left",
            y=1,
            yanchor="top",
            bgcolor="rgba(255,255,255,0.8)",  # Semi-transparent background
            bordercolor="black",
            borderwidth=1
        ),
        dragmode='pan',  # Default to pan mode
        autosize=True,  # Make it responsive to container size
        xaxis=dict(
            fixedrange=False,  # Allow zooming
            constrain='domain'  # Constrain to data domain
        ),
        yaxis=dict(
            fixedrange=False,  # Allow zooming
            constrain='domain'  # Constrain to data domain
        )
    )

    return fig
