from argparse import ArgumentParser, RawDescriptionHelpFormatter
import hashlib
import json
from pathlib import Path
import pickle
import os
import shlex
import subprocess
from io import StringIO
from tqdm import tqdm
import pandas as pd
from typing import Mapping
import plotly.express as px

from destine_analytics.lsreport import read_filetree
from .utils import sanitize_output

df = None


def _get_parser():
    parser = ArgumentParser(
        description="Parser for output of `tools/lstraverse.sh`.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "path",
        type=str,
        help=(
            "If it is a directory, local (or remote) path to draw filetree from. "
            "If it is a file, it is assumed to be the output of tools/lstraverse.sh."
        )
    )
    parser.add_argument(
        "-e", "--ssh-command",
        type=str,
        help=(
            "For remote path scoping: ssh command to execute in order to access to remote shell."
        )
    )
    parser.add_argument(
        "-E", "--expid",
        type=str,
        default=None,
        help="Expid. Only for title & filename purposes",
    )
    parser.add_argument(
        "-r", "--root",
        type=str,
        default="",
        help="Root directory to display. Must be within the file tree examined.",
    )
    parser.add_argument(
        "-d", "--only-directories",
        action="store_true",
        help="Only show directories in the figure. Needed for excessively large file trees.",
    )
    parser.add_argument(
        "-f", "--figure",
        type=str,
        default=None,
        choices=["treemap", "sunburst", "icicle", "all", None],
        help="Figure type to generate.",
    )
    parser.add_argument(
        "-i", "--interactive",
        action="store_true",
        help="Whether to interactively show the figure(s) after plotting.",
    )
    parser.add_argument(
        "-c", "--cache",
        action="store_true",
        help="Read / write to ~/.cache",
    )
    parser.add_argument(
        "-o", "--output",
        nargs="?",
        type=str,
        default=None,
        const="__default__",
        help=(
            "Output directory or file for saving figures. If no path is given, "
            "defaults to 'figures/'. If flag is not used, figure is not saved."
        )
    )
    return parser


def _json_serialize(obj):
    if isinstance(obj, Path):
        return str(obj)
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


def _get_cache_path(name, *args, **kwargs) -> Path:
    key = json.dumps([name, *args, kwargs], default=_json_serialize)
    filename = hashlib.sha256(key.encode("utf-8")).hexdigest()
    directory = os.path.basename(__file__)
    return Path.home().resolve() / ".cache" / directory / filename


def _cache_read(cache, name, /, *args, **kwargs):
    if cache is False:
        return
    cpath = _get_cache_path(name, *args, **kwargs)
    if not cpath.exists():
        return None
    print(f"Reading '{name}' from cache...")
    with open(cpath, "rb") as fb:
        return pickle.load(fb)


def _cache_write(cache, obj, name, /, *args, **kwargs):
    if cache is False:
        return
    print(f"Writing '{name}' to cache...")
    cpath = _get_cache_path(name, *args, **kwargs)
    if not cpath.parent.exists():
        cpath.parent.mkdir(parents=True)
    with open(cpath, "wb") as fb:
        pickle.dump(obj, fb)


def _add_path_column(df):
    # For lstraverse.sh output, Parent already contains the full directory path
    # We don't need to normalize it further, just create the full file paths
    df["Path"] = (df["Parent"] + "/" + df["Filename"]).apply(
        lambda x: x.removeprefix("/")
    )
    return df


def _get_subtree_df(df, root):
    # For lstraverse.sh output, Parent contains full directory paths
    df = df.copy()

    # Normalize requested root
    if root in (None, ""):
        # No root specified, return all data
        subtree_df = df.copy()
    else:
        # If root is absolute, use it directly
        if os.path.isabs(root):
            abs_root = os.path.normpath(root)
        else:
            # If root is relative, we need to find the common root first
            common_root = os.path.commonpath(list(df["Parent"]))
            abs_root = os.path.normpath(os.path.join(common_root, root))

        # Filter to subtree - only include files that are under the absolute root
        mask = df["Parent"].str.startswith(abs_root)
        subtree_df = df[mask].copy()

        if subtree_df.empty:
            raise ValueError(
                f"No entries found under root '{root}' (resolved to '{abs_root}')."
            )

    subtree_df["name"] = subtree_df["Filename"]

    return subtree_df


def _human_readable_size(size_bytes: int) -> str:
    """Convert size in bytes to human readable format with ~3 significant digits."""
    if size_bytes == 0:
        return "0B"
    for unit in ("B", "K", "M", "G", "T", "P", "E", "Z"):
        if abs(size_bytes) < 1024.0:
            if unit == "B":
                return f"{size_bytes:.0f}{unit}"
            if size_bytes < 10:
                return f"{size_bytes:.2f}{unit}"
            if size_bytes < 100:
                return f"{size_bytes:.1f}{unit}"
            return f"{size_bytes:.0f}{unit}"
        size_bytes /= 1024.0
    return f"{size_bytes:.2f}Y"


def draw(
    sizes: Mapping[str, int],
    parents: Mapping[str, str],
    root: str = None,
    figure: str = "treemap",
    expid: str = None,
):
    """
    Visualize file structure under a given root directory using Plotly.

    Parameters:
        sizes (Mapping[str, int]): Dictionary mapping directory paths to their total sizes
        parents (Mapping[str, str]): Dictionary mapping directory paths to their parent directories
        root (str | None): Subdirectory to start visualization from. Can be absolute or relative to the common root.
        figure (str): One of 'treemap', 'sunburst', or 'icicle'.
    """
    if figure not in {"treemap", "sunburst", "icicle"}:
        raise ValueError(
            f"Unsupported figure '{figure}'. Use 'treemap', 'sunburst', or 'icicle'."
        )

    # Create visualization DataFrame
    viz_df = pd.DataFrame(
        {
            "id": list(sizes.keys()),  # Full path as ID
            "parent": list(parents.values()),  # Full parent path
            "value": list(sizes.values()),  # Directory size
        }
    )

    # Add display name (last part of path + size)
    viz_df["name"] = viz_df.apply(
        lambda row: f"{row['id'].split('/')[-1]} ({_human_readable_size(row['value'])})",
        axis=1,
    )

    # Add human readable size for hover
    viz_df["size_str"] = viz_df["value"].apply(_human_readable_size)

    # Filter by root if specified
    if root:
        mask = viz_df["id"].str.startswith(root)
        viz_df = viz_df[mask]
        if viz_df.empty:
            raise ValueError(f"No entries found under root '{root}'.")

    # Create the figure
    fig = getattr(px, figure)(
        viz_df,
        ids="id",  # Use full paths as IDs
        names="name",  # Show name with size
        parents="parent",  # Use full parent paths
        values="value",
        hover_data=[
            "id",
            "size_str",
        ],  # Show full path and human readable size in hover
    )

    # Get the root for the title
    if root:
        abs_root = root
    else:
        # Try to find the root element (one with empty parent)
        root_entries = viz_df[viz_df["parent"] == ""]
        if not root_entries.empty:
            abs_root = root_entries["id"].iloc[0]
        else:
            # Fallback: use the first entry or a default
            if not viz_df.empty:
                abs_root = viz_df["id"].iloc[0]
            else:
                abs_root = "unknown"

    # Update traces for better visibility
    fig.update_traces(
        root_color="lightgrey",
        textposition="middle center",  # Center text in each box
        textfont=dict(size=14),  # Larger text
        textinfo="label",  # Show the label (name) in each box
    )

    # Update layout for better visibility and larger size
    title_text = f"{figure.title()} of directory sizes under: {abs_root}"
    if expid:
        title_text = f"[{expid}] {title_text}"
    fig.update_layout(
        title=dict(
            text=title_text,
            font=dict(size=24),  # Larger title
        ),
        margin=dict(t=60, l=0, r=0, b=0),  # More space for title
        width=2000,  # Wider figure
        height=1500,  # Taller figure
        font=dict(size=14),  # Larger font for all text
    )

    return fig


def makefigure(
    sizes: Mapping[str, int],
    parents: Mapping[str, str],
    root: str,
    figure: str,
    interactive: bool,
    output_dir: str,
    prefix: str,
    cache: bool,
    **kwargs,
):
    if figure == "all":
        return [
            makefigure(
                sizes, parents, root, "treemap", interactive, output_dir, prefix, cache, **kwargs
            ),
            makefigure(
                sizes, parents, root, "sunburst", interactive, output_dir, prefix, cache, **kwargs
            ),
            makefigure(
                sizes, parents, root, "icicle", interactive, output_dir, prefix, cache, **kwargs
            ),
        ]

    print(f"Plotting {figure}.")
    fig = _cache_read(cache, "figure", root=root, figure=figure, **kwargs)
    if fig is None:
        fig = draw(sizes, parents, root, figure, kwargs.get("expid"))
        _cache_write(cache, fig, "figure", root=root, figure=figure, **kwargs)

    filename = f"{prefix}{figure}.png"
    save_path = sanitize_output(output_dir, filename)
    if save_path:
        save_path.parent.mkdir(parents=True, exist_ok=True)
        fig.write_image(save_path)
    if interactive:
        print("Showing figure...")
        fig.show()

    return fig


def compute_directory_size(df, only_directories=False):
    """
    Compute directory sizes and their hierarchical relationships.

    Parameters:
        df (pd.DataFrame): DataFrame with 'Parent' and 'Filename' columns
        only_directories (bool): If True, only show directories in the visualization

    Returns:
        tuple: (sizes, parents) where:
            - sizes: dict mapping paths to their total sizes
            - parents: dict mapping paths to their parent directories
    """
    if df.empty:
        return {}, {}
    
    # Initialize dictionaries
    direct_sizes = {}  # Size of files directly in each directory
    total_sizes = {}  # Total size including subdirectories
    parents = {}

    # First pass: compute direct sizes and parent relationships
    # Start with directories
    for directory in tqdm(sorted(df["Parent"].unique())):
        # Get files directly in this directory
        mask = df["Parent"] == directory
        direct_sizes[directory] = int(df.loc[mask, "Filesize"].sum())
        total_sizes[directory] = direct_sizes[directory]  # Initialize total size

        # Set parent relationship based on actual directory structure
        if directory == "/" or directory == "":
            parents[directory] = ""
        else:
            # Find the parent directory by going up one level
            parent_dir = os.path.dirname(directory)
            if parent_dir == directory:  # Root directory
                parents[directory] = ""
            else:
                parents[directory] = parent_dir

    # Add files if not only_directories
    if not only_directories:
        for _, row in tqdm(df.iterrows(), total=len(df)):
            file_path = f"{row['Parent']}/{row['Filename']}"
            if (
                file_path not in direct_sizes
            ):  # Only add if not already added as a directory
                direct_sizes[file_path] = int(row["Filesize"])
                total_sizes[file_path] = direct_sizes[file_path]
                parents[file_path] = row["Parent"]

    # Second pass: propagate sizes up the tree
    # Process paths from longest to shortest
    for path in sorted(direct_sizes.keys(), key=len, reverse=True):
        parent = parents[path]
        if parent and parent in total_sizes:  # If not root and parent exists
            total_sizes[parent] += total_sizes[path]

    # Ensure we have at least one root element
    if not any(parent == "" for parent in parents.values()):
        # Find the shortest path as root
        if direct_sizes:
            shortest_path = min(direct_sizes.keys(), key=len)
            parents[shortest_path] = ""

    return total_sizes, parents

def _get_dataframe(path_str, ssh_command, expid, cache):
    if ssh_command:
        find_cmd = f"find {shlex.quote(path_str)} -ls"
        ssh_cmd_list = shlex.split(ssh_command) + [find_cmd]
        try:
            process = subprocess.run(
                ssh_cmd_list,
                capture_output=True,
                text=True,
                check=True,
                encoding="utf-8",
            )
            return read_filetree(StringIO(process.stdout), common=False)
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            raise RuntimeError(f"Error executing remote command: {e}") from e
    else:
        path = Path(path_str)
        if not path.exists():
            raise FileNotFoundError(f"Local path '{path}' does not exist.")

        if path.is_dir():
            print(f"Scanning directory '{path}'...")
            try:
                process = subprocess.run(
                    ["find", str(path), "-ls"],
                    capture_output=True,
                    text=True,
                    check=True,
                    encoding="utf-8",
                )
                return read_filetree(StringIO(process.stdout), common=False)
            except subprocess.CalledProcessError as e:
                raise FileNotFoundError(f"Error executing 'find' on '{path}': {e}") from e
        if path.is_file():
            return read_filetree(path, common=False)
        raise FileNotFoundError(f"Local path '{path}' is not a directory or a file.")


def main():
    global df
    args = _get_parser().parse_args()
    path_str = args.path
    ssh_command = args.ssh_command
    expid = args.expid
    root = args.root
    only_directories = args.only_directories
    figure = args.figure
    interactive = args.interactive
    cache = args.cache
    output_dir = args.output
    df = None


    cache_key_args = {
        "path": path_str,
        "expid": expid,
        "ssh_command": ssh_command,
    }
    df = _cache_read(cache, "df", **cache_key_args)
    if df is None:
        df = _get_dataframe(path_str, ssh_command, expid, cache)
        _cache_write(cache, df, "df", **cache_key_args)

    df = _add_path_column(df)
    df = _get_subtree_df(df, root)

    kwargs = {
        "path": path_str,
        "expid": expid,
        "root": root,
        "only_directories": only_directories,
        "ssh_command": ssh_command,
    }
    sizes_parents = _cache_read(cache, "sizes+parents", **kwargs)
    if sizes_parents is None:
        sizes, parents = compute_directory_size(df, only_directories=only_directories)
        sizes_parents = (sizes, parents)
        _cache_write(cache, sizes_parents, "sizes+parents", **kwargs)

    if figure is not None:
        if interactive and figure == "all":
            raise ValueError("Interactive mode cannot be used with --figure all. Please select a single figure type.")
        makefigure(
            sizes=sizes_parents[0],
            parents=sizes_parents[1],
            figure=figure,
            interactive=interactive,
            output_dir=output_dir,
            prefix=(f"{expid}_" if expid else ""),
            cache=cache,
            **kwargs,
        )


if __name__ == "__main__":
    main()
