import hashlib
import json
from pathlib import Path
import pickle
import os

from destine_analytics.paths import LOG_DIR

def get_available_expids() -> list[str]:
    """Get all available experiment IDs in the given log directory."""
    return [p.name for p in LOG_DIR.glob("*") if p.is_dir()]


def json_serialize(obj):
    if isinstance(obj, Path):
        return str(obj)
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


def get_cache_path(name, *args, **kwargs) -> Path:
    key = json.dumps([name, *args, kwargs], default=json_serialize)
    filename = hashlib.sha256(key.encode("utf-8")).hexdigest()
    directory = os.path.basename(__file__)
    return Path.home().resolve() / ".cache" / directory / filename


def cache_read(cache, name, /, *args, **kwargs):
    if cache is False:
        return
    cpath = get_cache_path(name, *args, **kwargs)
    if not cpath.exists():
        return None
    print(f"Reading '{name}' from cache...")
    with open(cpath, "rb") as fb:
        return pickle.load(fb)


def cache_write(cache, obj, name, /, *args, **kwargs):
    if cache is False:
        return
    print(f"Writing '{name}' to cache...")
    cpath = get_cache_path(name, *args, **kwargs)
    if not cpath.parent.exists():
        cpath.parent.mkdir(parents=True)
    with open(cpath, "wb") as fb:
        pickle.dump(obj, fb)

def sanitize_output(output_arg: str | None, filename: str) -> Path | None:
    """
    Sanitizes the output argument to produce a final save path.

    - If output_arg is None, returns None (don't save).
    - If output_arg is a special const, defaults to "figures/filename".
    - If output_arg is a directory, returns "output_arg/filename".
    - If output_arg is a file, returns it.
    - If output_arg is a bare filename, defaults to "figures/output_arg"

    Args:
        output_arg (str | None): The value from argparse.
        filename (str): The default filename for the plot.

    Returns:
        Path | None: The final path to save the figure, or None.
    """
    if output_arg is None:
        return None

    for default_dir in (Path.home() / "Pictures", Path.home() / "Desktop", Path.home() / "Documents", Path.home()):
        if default_dir.exists():
            break
    else:
        raise FileNotFoundError("No default directory found. Please provide an output directory.")

    # Special value for --output flag without argument
    if output_arg == "__default__":
        return default_dir / filename

    output_path = Path(output_arg)
    # If it's a dir, or has no suffix (assume dir), append filename
    if output_path.is_dir() or not output_path.suffix:
        return output_path / filename

    # Assumed to be a file path.
    if output_path.parent == Path("."):
        # It's a bare filename. Put it in the default directory.
        return default_dir / output_path

    # It's a full file path. Use it.
    return output_path